--------------------------------------------------------------------------------
--
--  Mod Name:		fieldStatus
--
--  Description:	Show status of all fields on a map (cultivated, ploughed,
--				etc).  Also show crops on the field, their growth states,
--				and if the field is fertilized.
--
--  Author:		Alan R. (ls-uk.info: thebadtouch)
--
--  License:		http://creativecommons.org/licenses/by-nc-sa/3.0/
--
--  Release History:
--  2011-07-24	v1.0		Initial release
--
--------------------------------------------------------------------------------
print("Script loading: fieldStatus.lua");

fieldStatus={};
fieldStatus.moddir=g_currentModDirectory;

local fileNew, mapName, saveTerrainDetailId, thisMap, thisMapId, thisPath, thisXMLFile = nil, nil, nil, nil, nil, nil, nil;
local debug, display, hasData = false, false, false;
local anyFieldsScanned, errMsgEnabled, isPrinted, isReloading, isScanning = false, false, false, false, false;
local loadSuccess, menuIsActive, savexml, userInputIsActive = false, false, false, false;
local bottomOK, nextOK, prevOK, topOK = false, false, false, false;
local ctrlSOK, ctrlROK, ctrlLOK = false, false, false;
local ctrlDOK, ctrlPOK, ctrlXOK = false, false, false;
local ctrlTOK = false;  local ctrlWOK = false;

local chname = {}; dispData = {}; fieldData = {}; fruitData = {}; holdData = {}; save_d = {}; scanData = {}; scanFieldArray = {};
local disp_d, dispDataItems, save_d_idx, start_d = 0, 0, 0, 0;
local maxCoords, numFields, numFruits = 0, 0, 0;
local numScanFields, scanIdx, scanNotMatch, scanTimer, searchIdx = 0, 0, 0, 0, 0;
local scanType = "full";
local dispColor, msgTxt, scanErrList, scanField, xmlLoadedFrom = "", "", "", "", "";
local mapHectares = 0;
local mapAcres = 0;

local displayTotalLines = 23;
local dispLinesAvail = 16;
----local lineSpacing = 0.021;
local lineSpacing = 0.02;
-- Old values: fsHudPosX = 0.813; fsHudPosY = 0.385; fsHudWidth = 0.176; fsHudHeight = 0.5;
-- New values: fsHudPosX = 0.778; fsHudPosY = 0.387; fsHudWidth = 0.211; fsHudHeight = 0.505;
local fsHudPosX = 0.789;
local fsHudPosY = 0.388;
local fsHudWidth = 0.211;
local fsHudHeight = 0.505;
local fsHudMargin = 0.007;
local fsPrintPosX = fsHudPosX + fsHudMargin;
local fsPrintPosXCenter = fsHudPosX + (fsHudWidth / 2);
local fsPrintPosXRight = fsHudPosX + fsHudWidth - fsHudMargin;
local fsPrintPosY = 0.83;		-- Old value: fsPrintPosY = 0.81;
local fsPrintMenuStart = fsPrintPosY - (lineSpacing * dispLinesAvail);
local fsPrintMenuLine1 = fsPrintMenuStart - lineSpacing;
local fsPrintMenuLine2 = fsPrintMenuStart - (lineSpacing * 2);
local fsPrintMenuLine3 = fsPrintMenuStart - (lineSpacing * 3);
local fsPrintMenuLine4 = fsPrintMenuStart - (lineSpacing * 4);
local fsPrintMenuLine5 = fsPrintMenuStart - (lineSpacing * 5);
local fsPrintMsgLine = fsPrintMenuStart - (lineSpacing * 6);
local textHeaderSize = 0.03;
local textNormalSize = 0.02;
local textMenuSize = 0.016;

local scanTimerWait = 250;
local scanTimerWait2 = 750;
local areaToHectare = 0.0001;
local hectareToAcre = 2.471;
local densityFactor = 4.41;
local maxCoordsDefault = 16;
local numFieldsDefault = 32;
local numGrowths = 9;
local numStatuses = 4;
local numdispDataEntries = 4;
local numholdDataEntries = 8;
local indent = "    ";
local fieldDataMaster = "fieldData.master.xml";
local fieldDataFile = "fieldData.xml";
local dash75 = "---------------------------------------------------------------------------";
local dash100 = "----------------------------------------------------------------------------------------------------";
local menuDashLine = "--------------------------------------";
local getdenFunc = Utils.getDensity;
local getdenregionFunc = getDensityRegion;
local getfruitFunc = Utils.getFruitArea;
local mfloor = math.floor;


function fieldStatus:loadMap(name)

if string.find(name, "/map01.i3d") ~= nil then
	local path, n = string.gsub(name, "/map01.i3d", "");
	thisMap, n = string.gsub(path, ".*/", "");
	if thisMap == "maps" then thisMap = "Default"; end;
	print("fieldStatus: thisMap >"..thisMap.."<");
end;

if g_currentMission.terrainDetailId ~= nil then
	if saveTerrainDetailId == nil then
		fieldStatus:load_data();
		if not loadSuccess then
			if xmlLoadedFrom == "savegame" then
				isReloading = true; fieldStatus:load_data(); isReloading = false;
				start_d = 0;
				if loadSuccess then
					msgTxt = "XML file loaded."
				end;
			end;
		else
			fieldStatus:process_fieldData("display");
		end;
	end;
end;

end;


function fieldStatus:deleteMap()

fileNew, mapName, saveTerrainDetailId, thisMap, thisMapId, thisPath, thisXMLFile = nil, nil, nil, nil, nil, nil, nil;
debug, display, hasData = false, false, false;
anyFieldsScanned, errMsgEnabled, isPrinted, isReloading, isScanning = false, false, false, false, false;
loadSuccess, menuIsActive, savexml, userInputIsActive = false, false, false, false;
bottomOK, nextOK, prevOK, topOK = false, false, false, false;
ctrlSOK, ctrlROK, ctrlLOK = false, false, false;
ctrlDOK, ctrlPOK, ctrlXOK = false, false, false;
ctrlTOK = false;  ctrlWOK = false;

chname = {}; dispData = {}; fieldData = {}; fruitData = {}; holdData = {}; save_d = {}; scanData = {}; scanFieldArray = {};
disp_d, dispDataItems, save_d_idx, start_d = 0, 0, 0, 0;
maxCoords, numFields, numFruits = 0, 0, 0;
numScanFields, scanIdx, scanNotMatch, scanTimer, searchIdx = 0, 0, 0, 0, 0;
scanType = "full";
dispColor, msgTxt, scanErrList, scanField, xmlLoadedFrom = "", "", "", "", "";
mapHectares = 0;
mapAcres = 0;

end;


function fieldStatus:mouseEvent(posX, posY, isDown, isUp, button)
end;


function fieldStatus:keyEvent(unicode, sym, modifier, isDown)

	if isDown then
		if bitAND(modifier,Input.MOD_CTRL) > 0 and (sym == Input.KEY_f) then
			if not loadSuccess then
				errMsgEnabled = not errMsgEnabled;
			else
				menuIsActive = not menuIsActive;
			end;
			if not menuIsActive then
				msgTxt = "";  dispColor = "";
			end;
		elseif bitAND(modifier,Input.MOD_CTRL) > 0 and (sym == Input.KEY_s) then
			if menuIsActive and ctrlSOK then
				msgTxt = "";  dispColor = "";
				isScanning = true;  scanType = "full";  scanTimer = 0;  scanIdx = 1;
			end;
		elseif bitAND(modifier,Input.MOD_CTRL) > 0 and (sym == Input.KEY_r) then
			if menuIsActive and ctrlROK then
				msgTxt = "";  dispColor = "";
				userInputIsActive = true; scanField = ""; sym = 0;
			end;
		elseif bitAND(modifier,Input.MOD_CTRL) > 0 and (sym == Input.KEY_l) then
			if menuIsActive and ctrlLOK then
				msgTxt = "";  dispColor = "";
				isReloading = true; fieldStatus:load_data(); isReloading = false;
				start_d = 0;
				if loadSuccess then
					msgTxt = "XML file loaded."
				end;
			end;
		elseif bitAND(modifier,Input.MOD_CTRL) > 0 and (sym == Input.KEY_d) then
			if menuIsActive and ctrlDOK then
				msgTxt = "";  dispColor = "";
				fieldStatus:process_fieldData("display");  start_d = 0;
			end;
		elseif bitAND(modifier,Input.MOD_CTRL) > 0 and (sym == Input.KEY_p) then
			if menuIsActive and ctrlPOK then
				msgTxt = "";  dispColor = "";
				fieldStatus:process_fieldData("debug");
			end;
		elseif bitAND(modifier,Input.MOD_CTRL) > 0 and (sym == Input.KEY_x) then
			if menuIsActive and ctrlXOK then
				msgTxt = "";  dispColor = "";
				fieldStatus:save_XML();
			end;
		elseif sym == Input.KEY_KP_7 then
			if menuIsActive and topOK then
				msgTxt = "";  dispColor = "";
				start_d = 0;
			end;
		elseif sym == Input.KEY_KP_1 then
			if menuIsActive and bottomOK then
				msgTxt = "";  dispColor = "";
				fieldStatus:findBottom();
			end;
		elseif sym == Input.KEY_KP_9 then
			if menuIsActive and prevOK then
				msgTxt = "";  dispColor = "";
				if save_d_idx > 1 then
					save_d[save_d_idx] = 0;
					save_d_idx = save_d_idx - 1;
					start_d = save_d[save_d_idx];
				end;
			end;
		elseif sym == Input.KEY_KP_3 then
			if menuIsActive and nextOK then
				msgTxt = "";  dispColor = "";
				start_d = disp_d;
			end;
		end;
		if menuIsActive and userInputIsActive then
			fieldStatus:getInput(sym);
			if not userInputIsActive then
				numScanFields = 0;
				if scanField ~= "" then
					scanField = scanField .. " ";
					scanFieldArray = {};
					local start = 1;
					local sfNew = string.upper(scanField);
					sfNew = string.gsub(sfNew, ",+", " ");
					sfNew = string.gsub(sfNew, " +", " ");
					sfNew = string.gsub(sfNew, "^ ", "");
					local a = string.find(sfNew, " ");
					while a do
						numScanFields = numScanFields + 1;
						scanFieldArray[numScanFields] = "Field " .. string.sub(sfNew, start, a - 1);
						start = a;
						sfNew = string.gsub(sfNew, " ", "", 1);
						a = string.find(sfNew, " ");
					end;
					if numScanFields > 0 then
						isScanning = true;  scanType = "part";  scanTimer = 0;  searchIdx = 1;  scanErrList = "";  anyFieldsScanned = false;
					end;
				end;
			end;
		end;
	end;
end;


function fieldStatus:update(dt)

	scanTimer = scanTimer + dt;

end;


function fieldStatus:draw()

if not loadSuccess then
	if errMsgEnabled then
		if scanTimer <= scanTimerWait2 then
			setTextAlignment(RenderText.ALIGN_CENTER);
			setTextBold(true);
			setTextColor(1, 0, 0, 1);				-- brightred
			renderText(0.5, 0.5, textHeaderSize, "fieldStatus: XML Error: Check log");
			renderText(0.5, 0.5 - textHeaderSize, textNormalSize, "(Ctrl-F to remove message)");
		elseif scanTimer > scanTimerWait2 * 2 then
			scanTimer = 0;
		end;
	end;
	return;
end;

if menuIsActive then
	renderOverlay(self.fsHUDId, fsHudPosX, fsHudPosY, fsHudWidth, fsHudHeight);
	setTextAlignment(RenderText.ALIGN_CENTER);
	setTextBold(true);
--	setTextColor(0.9, 0.18, 0.18, 1);	-- red
	setTextColor(0.11, 0.67, 0.16, 1); 	-- green
	renderText(fsPrintPosXCenter, fsPrintPosY + lineSpacing + fsHudMargin, textHeaderSize, "Field Status");
--	renderText(fsPrintPosX, fsPrintPosY + 0.04, 0.0245, "Field Status Menu");

	if isScanning then
		if scanType == "full" then
			msgTxt = "Scanning Field "..scanIdx.." of "..numFields.." ...";
			setTextAlignment(RenderText.ALIGN_LEFT);
			setTextBold(false);
			setTextColor(1, 1, 1, 1);				-- white
			renderText(fsPrintPosX, fsPrintMenuStart, textNormalSize, menuDashLine);
			setTextBold(true);
			setTextColor(0.31, 0.64, 0.93, 1);		-- blue
			renderText(fsPrintPosX, fsPrintMsgLine, textNormalSize, msgTxt);
			if scanTimer > scanTimerWait then
				scanNotMatch = 0;
				fieldStatus:scan_fields();
				scanTimer = 0;
				scanIdx = scanIdx + 1;
				if scanIdx > numFields or fieldData[scanIdx].name == nil then
					isScanning = false;
					msgTxt = "";  dispColor = "";
					fieldStatus:save_XML();
					if msgTxt == "" then
						msgTxt = "Scan Done.  XML file saved.";
					end;
					fieldStatus:process_fieldData("display");
					start_d = 0;
				end;
			end;
		else
			scanField = scanFieldArray[searchIdx];
			while true do
				scanIdx = 0;
				local scanStart;
				for scanStart=1, numFields do
					if fieldData[scanStart].name ~= nil and fieldData[scanStart].name == scanField then
						scanIdx = scanStart;
						break;
					end;
				end;
				if scanIdx == 0 then  -- Field not found
					scanErrList = scanErrList .. scanField .. ",";
					searchIdx = searchIdx + 1;
					if searchIdx > numScanFields then
						isScanning = false;
						break;
					else
						return;
					end;
				else
					break;
				end;
			end;
			if isScanning then
				msgTxt = "Scanning "..fieldData[scanIdx].name.." ...";
				setTextAlignment(RenderText.ALIGN_LEFT);
				setTextBold(false);
				setTextColor(1, 1, 1, 1);				-- white
				renderText(fsPrintPosX, fsPrintMenuStart, textNormalSize, menuDashLine);
				setTextBold(true);
				setTextColor(0.31, 0.64, 0.93, 1);		-- blue
				renderText(fsPrintPosX, fsPrintMsgLine, textNormalSize, msgTxt);
				if scanTimer > scanTimerWait then
					scanNotMatch = 0;
					fieldStatus:scan_fields();
					scanTimer = 0;
					searchIdx = searchIdx + 1;
					anyFieldsScanned = true;
				end;
			end;
			if searchIdx > numScanFields then
				isScanning = false;
				msgTxt = "";  dispColor = "";
				if anyFieldsScanned then
					fieldStatus:save_XML();
					if msgTxt == "" then
						msgTxt = "Scan Done.  XML file saved.";
					end;
					fieldStatus:process_fieldData("display");
				end;
				if scanErrList ~= "" then
					scanErrList = string.gsub(scanErrList, "Field ", "");
					scanErrList = string.sub(scanErrList, 1, -2);
					msgTxt = "No Fields: " .. scanErrList;
					dispColor = "brightred";
				end;
				scanField = "";  scanFieldArray = {};  numScanFields = 0;  scanErrList = "";
			end;
		end;
		return;
	elseif userInputIsActive then
		msgTxt = "Field(s) >"..scanField.."<";
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		setTextColor(1, 1, 1, 1);				-- white
		renderText(fsPrintPosX, fsPrintMenuStart, textNormalSize, menuDashLine);
		setTextColor(0.31, 0.64, 0.93, 1);		-- blue
		renderText(fsPrintPosX, fsPrintMsgLine, textNormalSize, msgTxt);
		return;
	end;
else
	return;
end;

local fieldDone, screenDone = false, false;
local i, linesPrinted, numFieldLines = 0, 0, 0;
local fsPrintPosCurrLine = fsPrintPosY;
local linesAvail = dispLinesAvail;
local blankTxt = " ";
local text1 = "";

local menu0a = "Tot. Area: ";					local menu0b = string.format("%4.1f ha / %4.1f ac", mapHectares, mapAcres);
local menu1a = "KP 7: Top";					local menu1b = "KP 9: Previous";
local menu2a = "KP 1: Bottom";					local menu2b = "KP 3: Next";
local menu3a = "Ctrl-S: Scan All";				local menu3b = "Ctrl-R: Scan Some";
local menu4a = "Ctrl-L: Reload XML";

topOK = true;  bottomOK = true;  prevOK = true;  nextOK = true;
ctrlSOK = true;  ctrlROK = true;  ctrlLOK = true;
ctrlDOK = true;  ctrlXOK = true;  ctrlPOK = true;

if start_d > 0 then
	disp_d = start_d;
	if disp_d > save_d[save_d_idx] then
		save_d_idx = save_d_idx + 1;
		save_d[save_d_idx] = disp_d;  -- save a new entry in the "prev" list
	end;
else
	disp_d = 1;
	start_d = disp_d;
	save_d = {};  save_d_idx = 1;  save_d[save_d_idx] = disp_d;
end;

if disp_d == 1 then
	menu1a = " ";  topOK = false;  menu1b = " ";  prevOK = false;
end;

for i=1,displayTotalLines do
	if (not fieldDone) and (not screenDone) then
		if dispData[disp_d].ctrl == nil then
			screenDone = true;
			if disp_d == 1 then
				blankTxt = " ";
			else
				blankTxt = "* END OF DATA *";
			end;
			menu2a = " ";  bottomOK = false;  menu2b = " "; nextOK = false;
		elseif not hasData then
			screenDone = true;
			blankTxt = "* NO DATA FOUND *";
			menu2a = " ";  bottomOK = false;  menu2b = " "; nextOK = false;
		else
			text1 = string.sub(dispData[disp_d].ctrl, 1, 3);
			if text1 == "FLD" then
				local e, s = string.find(dispData[disp_d].ctrl, "NUM");
				if e ~= nil then
					numFieldLines = tonumber(string.sub(dispData[disp_d].ctrl, s+1));
				else
					if not isPrinted then
						fieldStatus:showdispData();
						isPrinted = true;
					end;
					msgTxt = "ERR: Invalid data: See log";
					dispColor = "brightred";
					screenDone = true;
				end;
				if numFieldLines > linesAvail then  -- the next field contains too many lines to fit in the remaining lines
					screenDone = true;
				else
					fieldDone = false;
					linesPrinted = 0;
				end;
			elseif text1 == "EMP" then
----				setTextColor(1, 1, 1, 1);			-- white
----				setTextColor(0, 0, 0, 1);			-- black
				setTextColor(0.83, 0.74, 0.62, 1); 	-- brown (dirt)
			elseif text1 == "STA" then
				setTextColor(0.83, 0.74, 0.62, 1); 	-- brown (dirt)
			elseif text1 == "FER" then
				setTextColor(0.31, 0.64, 0.93, 1); 	-- blue
			elseif text1 == "FRU" then
				setTextColor(0.11, 0.67, 0.16, 1); 	-- green
			elseif text1 == "CUT" then
				setTextColor(0.78, 0.76, 0.16, 1); 	-- yellow
			end;
		end;
	end;

	if i == displayTotalLines - 6 then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		setTextColor(1, 1, 1, 1);				-- white
		renderText(fsPrintPosX, fsPrintMenuStart, textNormalSize, menuDashLine);
	elseif i == displayTotalLines - 5 then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		setTextColor(0.31, 0.64, 0.93, 1); 		-- blue
		renderText(fsPrintPosX, fsPrintMenuLine1, textNormalSize, menu0a);
		setTextAlignment(RenderText.ALIGN_RIGHT);
		setTextBold(false);
		setTextColor(0.31, 0.64, 0.93, 1); 		-- blue
		renderText(fsPrintPosXRight, fsPrintMenuLine1, textNormalSize, menu0b);
	elseif i == displayTotalLines - 4 then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		setTextColor(1, 1, 1, 1);				-- white
		renderText(fsPrintPosX, fsPrintMenuLine2, textMenuSize, menu1a);
		renderText(fsPrintPosXCenter, fsPrintMenuLine2, textMenuSize, menu1b);
	elseif i == displayTotalLines - 3 then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		setTextColor(1, 1, 1, 1);				-- white
		renderText(fsPrintPosX, fsPrintMenuLine3, textMenuSize, menu2a);
		renderText(fsPrintPosXCenter, fsPrintMenuLine3, textMenuSize, menu2b);
	elseif i == displayTotalLines - 2 then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		setTextColor(1, 1, 1, 1);				-- white
		renderText(fsPrintPosX, fsPrintMenuLine4, textMenuSize, menu3a);
		renderText(fsPrintPosXCenter, fsPrintMenuLine4, textMenuSize, menu3b);
	elseif i == displayTotalLines - 1 then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		setTextColor(1, 1, 1, 1);				-- white
		renderText(fsPrintPosX, fsPrintMenuLine5, textMenuSize, menu4a);
	elseif i == displayTotalLines then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		if dispColor == "brightred" then
			setTextColor(1, 0, 0, 1);			-- brightred
		else
			setTextColor(0.31, 0.64, 0.93, 1);	-- blue (default message color)
		end;
		renderText(fsPrintPosX, fsPrintMsgLine, textNormalSize, msgTxt);
	elseif screenDone or fieldDone then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(true);
		setTextColor(0.31, 0.64, 0.93, 1);		-- blue (default message color)
		renderText(fsPrintPosX, fsPrintPosCurrLine, textNormalSize, blankTxt);
		blankTxt = " ";
		if fieldDone then fieldDone = false; end;
	elseif text1 == "FLD" then
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(true);
		setTextColor(1, 1, 1, 1); 				-- white
		renderText(fsPrintPosX, fsPrintPosCurrLine, textNormalSize, dispData[disp_d].a);
		if dispData[disp_d].b ~= nil then                              -- Hectares / Acres
			setTextAlignment(RenderText.ALIGN_RIGHT);
			setTextBold(false);
			renderText(fsPrintPosXRight, fsPrintPosCurrLine, textNormalSize, dispData[disp_d].b);
		end;
		disp_d = disp_d + 1; linesPrinted = linesPrinted + 1;
	else
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextBold(false);
		renderText(fsPrintPosX, fsPrintPosCurrLine, textNormalSize, dispData[disp_d].a);
		setTextAlignment(RenderText.ALIGN_RIGHT);
		renderText(fsPrintPosXRight, fsPrintPosCurrLine, textNormalSize, dispData[disp_d].b);
		disp_d = disp_d + 1; linesPrinted = linesPrinted + 1;
	end;
	fsPrintPosCurrLine = fsPrintPosCurrLine - lineSpacing;
	linesAvail = linesAvail - 1;
	if linesAvail == 0 then
		screenDone=true;
	end;
	if linesPrinted == numFieldLines then
		if dispDataItems ~= disp_d - 1 then		-- Basically, if there are more fields to display, set fieldDone to true; otherwise, just let 'end of data' msg print (top of 'for' loop)
			fieldDone=true;
			linesPrinted = 0;
		end;
	end;
end;

end;


function fieldStatus:load_data()

if debug then print("fieldStatus: BEG func load_data"); end;

local k, m, n, o;
loadSuccess = false;
errMsgEnabled = false;
xmlLoadedFrom = "";
mapHectares = 0;
mapAcres = 0;

if not isReloading then
	saveTerrainDetailId = g_currentMission.terrainDetailId;
	numFruits = table.getn(g_currentMission.fruits);
	print("fieldStatus: terrainDetailId:      "..saveTerrainDetailId);
	print("fieldStatus: terrainSize:          "..g_currentMission.terrainSize);
	print("fieldStatus: terrainDetailMapSize: "..g_currentMission.terrainDetailMapSize);
	print("fieldStatus: numFruits:            "..numFruits);

	self.fsHUDPath = Utils.getFilename("fs_hud.png", fieldStatus.moddir);
	self.fsHUDId=createImageOverlay(self.fsHUDPath);
----	setOverlayColor(self.fsHUDId, 155, 150, 150, 1);

	print("fieldStatus: Loading data from xml file in savegame directory.");
	thisPath = getUserProfileAppPath().."savegame"..g_careerScreen.selectedIndex.."/";
	thisXMLFile = Utils.getFilename(fieldDataFile, thisPath);
		if debug then print("thisXMLFile >"..thisXMLFile.."<"); end;
	self.xmlFile = loadXMLFile("xmlFile", thisXMLFile);
		if debug then print("self.xmlFile >"..self.xmlFile.."<"); end;
	fileNew = io.open(thisXMLFile, "r");
	if fileNew == nil then
		print("fieldStatus: "..thisXMLFile.." not found...  Loading data from master xml file in mods directory.");
		thisPath = fieldStatus.moddir;
		thisXMLFile = Utils.getFilename(fieldDataMaster, thisPath);
		self.xmlFile = loadXMLFile("xmlFile", thisXMLFile);
		xmlLoadedFrom = "master";
	else
		fileNew:close();
		xmlLoadedFrom = "savegame";
	end;
else
	print("fieldStatus: Loading data from master xml file in mods directory.");
	thisPath = fieldStatus.moddir;
	thisXMLFile = Utils.getFilename(fieldDataMaster, thisPath);
	self.xmlFile = loadXMLFile("xmlFile", thisXMLFile);
	xmlLoadedFrom = "master";
end;

m=0;
while true do
	local basename=string.format("fieldData.map(%d)", m);
	mapName = getXMLString(self.xmlFile, basename.."#name");
	if mapName == nil then
		if m == 0 then
			print("fieldStatus: Error: No Maps found in: "..thisXMLFile);
		else
			print("fieldStatus: Error: "..thisMap.." not found in: "..thisXMLFile);
		end;
		loadSuccess=false;
		errMsgEnabled = true;
		break;
	elseif mapName == thisMap then
		thisMapId=m;
		print("fieldStatus: mapName >"..mapName.."<");
		numFields = Utils.getNoNil(getXMLInt(self.xmlFile, basename.."#numFields"), numFieldsDefault);
		maxCoords = Utils.getNoNil(getXMLInt(self.xmlFile, basename.."#maxCoords"), maxCoordsDefault);
		if numFields <= 0 then
			numFields = numFieldsDefault;
		end;
		if maxCoords <= 0 then
			maxCoords = maxCoordsDefault;
		end;
		loadSuccess=true;
		break;
	end;
	m=m+1;
end;

if not loadSuccess then
	return 99;
end;

fieldStatus:init_data();

chname[0]="0: cult";
chname[1]="1: plow";
chname[2]="2: seed";
chname[3]="3: fert";

	k=thisMapId; m=0;
	while true do
		local basename=string.format("fieldData.map(%d).field(%d)", k, m);
		local fieldName = getXMLString(self.xmlFile, basename.."#name");
		if fieldName == nil then
			if m == 0 then
				print("fieldStatus: Error: No Fields found for ", mapName);
				loadSuccess=false;
				errMsgEnabled = true;
			end;
			break;
		else
			if debug then print("    fieldName >"..fieldName.."<"); end;
			fieldData[m+1].name = fieldName;
			local fieldArea = getXMLInt(self.xmlFile, basename.."#area");
			if fieldArea == nil or fieldArea <= 0 then
				print("fieldStatus: Error: Missing Area for ", fieldName);
				loadSuccess=false;
				errMsgEnabled = true;
				break;
			else
				fieldData[m+1].area = fieldArea;
				local fieldHectares = getXMLFloat(self.xmlFile, basename.."#hectares");
				if fieldHectares == nil or fieldHectares <= 0 then
					fieldHectares = fieldArea * areaToHectare;
				end;
				local fieldAcres = getXMLFloat(self.xmlFile, basename.."#acres");
				if fieldAcres == nil or fieldAcres <= 0 then
					fieldAcres = fieldHectares * hectareToAcre;
				end;
				fieldData[m+1].hectares = fieldHectares;
				fieldData[m+1].acres = fieldAcres;
				mapHectares = mapHectares + fieldHectares;
				mapAcres = mapAcres + fieldAcres;
			end;
		end;
		n=0;
		while true do
			local basename=string.format("fieldData.map(%d).field(%d).coordinates(%d)", k, m, n);
			local c1 = getXMLInt(self.xmlFile, basename.."#startX");
			local c2 = getXMLInt(self.xmlFile, basename.."#startZ");
			local c3 = getXMLInt(self.xmlFile, basename.."#endX");
			local c4 = getXMLInt(self.xmlFile, basename.."#endZ");
			if c1 == nil or c2 == nil or c3 == nil or c4 == nil then
				if n == 0 then
					print("fieldStatus: Error: Missing Coordinates for ", fieldName);
					loadSuccess=false;
					errMsgEnabled = true;
				end;
				break;
			else
				if debug then print("        coordinates >"..c1..", "..c2..", "..c3..", "..c4.."<"); end;
				fieldData[m+1].coord[n+1].startX = c1;
				fieldData[m+1].coord[n+1].startZ = c2;
				fieldData[m+1].coord[n+1].endX = c3;
				fieldData[m+1].coord[n+1].endZ = c4;
			end;
			n=n+1;
			if n > maxCoords then
				break;
			end;
		end;
		if loadSuccess then
			n=0;
			while true do
				local basename=string.format("fieldData.map(%d).field(%d).status(%d)", k, m, n);
				local s1 = getXMLInt(self.xmlFile, basename.."#num");
				local s2 = getXMLFloat(self.xmlFile, basename.."#density");
				if s1 == nil or s2 == nil then
					break;
				else
					if debug then print("        status >"..s1.."  density: "..string.format("%9d", s2).."<"); end;
					fieldData[m+1].status[s1].num = s1;
					fieldData[m+1].status[s1].density = s2;
				end;
				n=n+1;
			end;
		end;
		if loadSuccess then
			n=0;
			while true do
				local basename=string.format("fieldData.map(%d).field(%d).fruit(%d)", k, m, n);
				local f1 = getXMLString(self.xmlFile, basename.."#name");
				if f1 == nil then
					break;
				else
					if debug then print("        fruit >"..f1.."<"); end;
					fieldData[m+1].fruit[n+1].name = f1;
				end;
				o=0;
				while true do
					local basename=string.format("fieldData.map(%d).field(%d).fruit(%d).growth(%d)", k, m, n, o);
					local g1 = getXMLInt(self.xmlFile, basename.."#num");
					local g2 = getXMLFloat(self.xmlFile, basename.."#density");
					if g1 == nil or g2 == nil then
						break;
					else
						if debug then print("            growth >"..g1.."  density: "..string.format("%9d", g2).."<"); end;
						fieldData[m+1].fruit[n+1].growth[g1].num = g1;
						fieldData[m+1].fruit[n+1].growth[g1].density = g2;
					end;
					o=o+1;
				end;
				n=n+1;
			end;
		end;
		m=m+1;
		if m > numFields then
			break;
		end;
	end;

if not loadSuccess then
	return 99;
end;

if debug then print("fieldStatus: END func load_data"); end;
return 0;
end;


function fieldStatus:init_data()

local m, n, o;
for m=1,numFields do
	fieldData[m] = {};			scanData[m] = {};
	fieldData[m].coord = {};
	fieldData[m].status = {};		scanData[m].status = {};
	fieldData[m].fruit = {};		scanData[m].fruit = {};
	for n=1,maxCoords do
		fieldData[m].coord[n] = {};
	end;
	for n=0,numStatuses-1 do
		fieldData[m].status[n] = {};		scanData[m].status[n] = {};
		fieldData[m].status[n].density = 0;	scanData[m].status[n].density = 0;
	end;
	for n=1,numFruits do
		fruitData[n] = {};
		fruitData[n].name = FruitUtil.fruitIndexToDesc[n].name;
		if g_currentMission.fruits[n] ~= nil then
			fruitData[n].id         = g_currentMission.fruits[n].id;
			fruitData[n].cutShortId = g_currentMission.fruits[n].cutShortId;
			fruitData[n].cutLongId  = g_currentMission.fruits[n].cutLongId;
			fruitData[n].windrowId  = g_currentMission.fruits[n].windrowId;
		end;
		fieldData[m].fruit[n] = {};		scanData[m].fruit[n] = {};
		fieldData[m].fruit[n].growth = {};	scanData[m].fruit[n].growth = {};
		for o=0,numGrowths-1 do
			fieldData[m].fruit[n].growth[o] = {};			scanData[m].fruit[n].growth[o] = {};
			fieldData[m].fruit[n].growth[o].density = 0;	scanData[m].fruit[n].growth[o].density = 0;
		end;
	end;
end;

for m=1,numFields*numGrowths*numdispDataEntries do
	dispData[m] = {};
end;
dispDataItems = 0;
for m=1,numholdDataEntries do
	holdData[m] = {};
end;

end;


function fieldStatus:reset_data(resIdx)

local m, n, o;
--for m=1,numFields do
	fieldData[resIdx].status = {};		scanData[resIdx].status = {};
	fieldData[resIdx].fruit = {};		scanData[resIdx].fruit = {};
	for n=0,numStatuses-1 do
		fieldData[resIdx].status[n] = {};		scanData[resIdx].status[n] = {};
		fieldData[resIdx].status[n].density = 0;	scanData[resIdx].status[n].density = 0;
	end;
	for n=1,FruitUtil.NUM_FRUITTYPES do
		fieldData[resIdx].fruit[n] = {};		scanData[resIdx].fruit[n] = {};
		fieldData[resIdx].fruit[n].growth = {};	scanData[resIdx].fruit[n].growth = {};
		for o=0,numGrowths-1 do
			fieldData[resIdx].fruit[n].growth[o] = {};			scanData[resIdx].fruit[n].growth[o] = {};
			fieldData[resIdx].fruit[n].growth[o].density = 0;	scanData[resIdx].fruit[n].growth[o].density = 0;
		end;
	end;
--end;

--for m=1,numFields*numGrowths*numdispDataEntries do
	dispData[resIdx] = {};
--end;
dispDataItems = 0;

end;


function fieldStatus:reset_scan_data(resIdx)

local m, n, o;
--for m=1,numFields do
	scanData[resIdx].status = {};
	scanData[resIdx].fruit = {};
	for n=0,numStatuses-1 do
		scanData[resIdx].status[n] = {};
		scanData[resIdx].status[n].density = 0;
	end;
	for n=1,FruitUtil.NUM_FRUITTYPES do
		scanData[resIdx].fruit[n] = {};
		scanData[resIdx].fruit[n].growth = {};
		for o=0,numGrowths-1 do
			scanData[resIdx].fruit[n].growth[o] = {};
			scanData[resIdx].fruit[n].growth[o].density = 0;
		end;
	end;
--end;

end;


function fieldStatus:getInput(sym)

	if sym == Input.KEY_return then							--  RETURN or ENTER
		userInputIsActive = false;
		msgTxt = "";  dispColor = "";
	elseif sym == 8 then									--  BACKSPACE
		if string.len(scanField) > 1 then
			scanField = string.sub(scanField, 1, -2);
		else
			scanField = "";
		end;
	elseif (sym >= 48) and (sym <= 57) then					--  0 - 9
		scanField = scanField .. string.char(sym);
	elseif (sym >= 65) and (sym <= 90) then					--  A - Z
		scanField = scanField .. string.char(sym);
	elseif (sym >= 97) and (sym <= 122) then					--  a - z
		scanField = scanField .. string.char(sym - 32);
	elseif sym == 32 or sym == 44 then						--  space or comma
		scanField = scanField .. string.char(sym);
	end;
end;


function fieldStatus:scan_fields()

if debug then print("fieldStatus: BEG func scan_fields @ "..os.date("%H:%M:%S").."    scanIdx: "..scanIdx);  end;
if debug then local xtime = os.clock();  end;

local r, t;
local this_row;
local isSeeded;
local thisfruit;
local startX, startZ, endX, endZ, widthX, widthZ, heightX, heightZ;
local i, m, n, o, k, v, x, z, xx, zz, xrem, zrem, growthX, growthZ;
local chnum, density, growth, saveDensity;

local widthXAdj=10;
local widthZAdj=0;
local heightXAdj=0;
local heightZAdj=10;

if debug then print("scan values:    width: "..widthXAdj..", "..widthZAdj.."  height: "..heightXAdj..", "..heightZAdj); end;

m = scanIdx;
fieldStatus:reset_data(m)

-- for m=1,numFields do

if fieldData[m].name ~= nil then

for n=1,maxCoords do
	if fieldData[m].coord[n].startX ~= nil then
		startX=(fieldData[m].coord[n].startX-1024); startZ=(fieldData[m].coord[n].startZ-1024);
		endX=(fieldData[m].coord[n].endX-1024); endZ=(fieldData[m].coord[n].endZ-1024);

--		local widthWorldX, widthWorldZ, heightWorldX, heightWorldZ;
--		widthWorldX=1; widthWorldZ=1; heightWorldX=0; heightWorldZ=3;
--		if debug then print("widthWorldX: "..widthWorldX.." widthWorldZ: "..widthWorldZ.." heightWorldX: "..heightWorldX.." heightWorldZ: "..heightWorldZ); end;
--		x, z, widthX, widthZ, heightX, heightZ = Utils.getXZWidthAndHeight(saveTerrainDetailId, startX, startZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ); 
--		if debug then print("x,z: "..x..","..z.."  widthX,widthZ: "..widthX..","..widthZ.."  heightX,heightZ: "..heightX..","..heightZ); end;

	if debug then print("estimated cols:  "..tonumber((endX-startX)/widthXAdj)); end;
	if debug then print("estimated rows:  "..tonumber((endZ-startZ)/heightZAdj)); end;
	xrem = (endX-startX) % widthXAdj;
	zrem = (endZ-startZ) % heightZAdj;
	if debug then print("remainders:      "..xrem..", "..zrem); end;

	this_row=0;

	for z=startZ,endZ-1,heightZAdj do
		this_row=this_row+1;
		if debug then print("row: "..this_row); end;
			if z > endZ-heightZAdj then
				heightZ=z+zrem; growthZ=(zrem*2);
					if debug then print("on last row:   z,endZ: "..z..","..endZ.."  rem: "..zrem.."  heightZ: "..heightZ.."  growthZ: "..growthZ); end;
			else
				heightZ=z+heightZAdj; growthZ=20;
			end;
	for x=startX,endX-1,widthXAdj do
			if x > endX-widthXAdj then
				widthX=x+xrem; growthX=(xrem*2);
					if debug then print("on last col:   x,endX: "..x..","..endX.."  rem: "..xrem.."  widthX: "..widthX.."  growthX: "..growthX); end;
			else
				widthX=x+widthXAdj; growthX=20;
			end;

	local scanfruits = 0;
	widthZ=z+widthZAdj; heightX=x+heightXAdj;

	if debug then print(dash100); end;
	if debug then print(fieldData[m].name.."    x,z: "..x..","..z.."  widthX,widthZ: "..widthX..","..widthZ.."  heightX,heightZ: "..heightX..","..heightZ); end;

	isSeeded=false;  saveDensity = 0;
	for chnum=0,3 do
		density = getdenFunc(saveTerrainDetailId, chnum, x, z, widthX, widthZ, heightX, heightZ);
		if density > 0 then
			fieldData[m].status[chnum].density = fieldData[m].status[chnum].density + density;
			if debug then print("terrain: "..saveTerrainDetailId.."  chan: "..chname[chnum].."  density: "..string.format("%6d", density)); end;
			if chnum == 2 then
				isSeeded=true;  saveDensity = density;
			end;
		end;
	end;

	for i=1,numFruits do
	if fruitData[i] ~= nil then
		local fruitPrinted = false;
		thisfruit = fruitData[i].name;
--		if debug then if not fruitPrinted then fruitPrinted = true; print(dash75); end; end;
--		if debug then print("i: "..i.."  id: "..fruitData[i].id.."  cutShortId: "..fruitData[i].cutShortId.."  cutLongId: "..fruitData[i].cutLongId.."  windrowId: "..fruitData[i].windrowId); end;

		chnum=2;  -- check for growing/grown fruit only on sowingChannel
		if fruitData[i].id ~= nil and fruitData[i].id > 0 then
--			density = getdenFunc(fruitData[i].id, chnum, x, z, widthX, widthZ, heightX, heightZ);
--				if density > 0 then
--					if debug then if not fruitPrinted then fruitPrinted = true; print(dash75); end; end;
--					if debug then print("D "..string.format("%26s", thisfruit).."  density: "..string.format("%6d", density)); end;
--				end;
			density = getfruitFunc(FruitUtil.fruitIndexToDesc[i].index, x, z, widthX, widthZ, heightX, heightZ);
			if density > 0 then
				if debug then if not fruitPrinted then fruitPrinted = true; print(dash75); end; end;
				if debug then print("F "..string.format("%26s", thisfruit).."  density: "..string.format("%6d", density)); end;
				local growthFound = false;
				xx=(x+1024)*2; zz=(z+1024)*2;
---				if debug then print("xx,zz: "..xx..","..zz.."  widthXAdj,widthZAdj: "..widthXAdj..","..widthZAdj.."  heightXAdj,heightZAdj: "..heightXAdj..","..heightZAdj); end;
				local ww, yy;
				for yy = zz, zz+growthZ-2, 2 do
					for ww = xx, xx+growthX-2, 2 do
---						if debug then print("fruit G: ww, yy: "..ww..", "..yy); end;
						growth = getdenregionFunc(fruitData[i].id, ww, yy, 1, 1, 0, 3);
						if growth > 0 then
----							fieldData[m].fruit[i].name = thisfruit; fieldData[m].fruit[i].growth[growth].density = fieldData[m].fruit[i].growth[growth].density + density;
----							fieldData[m].fruit[i].growth[growth].num = growth;
							scanData[m].fruit[i].name = thisfruit; scanData[m].fruit[i].growth[growth].density = density;
							scanData[m].fruit[i].growth[growth].num = growth;
							scanfruits = scanfruits + 1;
							if debug then print("G "..string.format("%26s", thisfruit)..string.format("%27s%2d", "growth: ", growth)); end;
							growthFound = true;  break;
						end;
					end;
					if growthFound then break;  end;
				end;
				if not growthFound then
					scanData[m].fruit[i].name = thisfruit; scanData[m].fruit[i].growth[0].density = density;
					scanData[m].fruit[i].growth[0].num = 0;
					scanfruits = scanfruits + 1;
--					for t=z,heightZ-1,5 do
--						for r=x,widthX-1,5 do
--							density = getfruitFunc(FruitUtil.fruitIndexToDesc[i].index, r, t, 5, 0, 0, 5);
--							if density > 0 then
--								if debug then print("XX"..string.format("%26s", thisfruit).."  density: "..string.format("%6d", density)); end;
--							end;
--						end;
--					end;
				end;
			elseif isSeeded then
				local growthFound = false;
				xx=(x+1024)*2; zz=(z+1024)*2;
				local ww, yy;
				for yy = zz, zz+growthZ-2, 2 do
					for ww = xx, xx+growthX-2, 2 do
---						if debug then print("fruit 5: ww, yy: "..ww..", "..yy); end;
						growth = getdenregionFunc(fruitData[i].id, ww, yy, 1, 1, 0, 3);
						if growth > 0 then
----							fieldData[m].fruit[i].name = thisfruit; fieldData[m].fruit[i].growth[growth].density = fieldData[m].fruit[i].growth[growth].density + density;
----							fieldData[m].fruit[i].growth[growth].num = growth;
							scanData[m].fruit[i].name = thisfruit; scanData[m].fruit[i].growth[growth].density = density;
							scanData[m].fruit[i].growth[growth].num = growth;
							scanfruits = scanfruits + 1;
							if debug then print("NS"..string.format("%26s", thisfruit)..string.format("%27s%2d", "growth: ", growth)); end;
							growthFound = true;  break;
						end;
					end;
					if growthFound then break;  end;
				end;
			end;

		end;  -- fruitData id not nil and > 0

		chnum=0;  -- check for cut/windrow fruit only on cultivatorChannel
			if fruitData[i].cutShortId ~= nil and fruitData[i].cutShortId > 0 then
				density = getdenFunc(fruitData[i].cutShortId, chnum, x, z, widthX, widthZ, heightX, heightZ);
				if density > 0 then
					if debug then if not fruitPrinted then fruitPrinted = true; print(dash75); end; end;
					if debug then print(string.format("%28s", thisfruit).."  density: "..string.format("%6d", density).."  cutshort"); end;
					fieldData[m].fruit[i].name = thisfruit; fieldData[m].fruit[i].growth[6].density = fieldData[m].fruit[i].growth[6].density + density;
					fieldData[m].fruit[i].growth[6].num = 6;
				end;
			end;
			if fruitData[i].cutLongId ~= nil and fruitData[i].cutLongId > 0 then
				density = getdenFunc(fruitData[i].cutLongId, chnum, x, z, widthX, widthZ, heightX, heightZ);
				if density > 0 then
					if debug then if not fruitPrinted then fruitPrinted = true; print(dash75); end; end;
					if debug then print(string.format("%28s", thisfruit).."  density: "..string.format("%6d", density).."  cutlong"); end;
					fieldData[m].fruit[i].name = thisfruit; fieldData[m].fruit[i].growth[7].density = fieldData[m].fruit[i].growth[7].density + density;
					fieldData[m].fruit[i].growth[7].num = 7;
				end;
			end;
			if fruitData[i].windrowId ~= nil and fruitData[i].windrowId > 0 then
				density = getdenFunc(fruitData[i].windrowId, chnum, x, z, widthX, widthZ, heightX, heightZ);
				if density > 0 then
					if debug then if not fruitPrinted then fruitPrinted = true; print(dash75); end; end;
					if debug then print(string.format("%28s", thisfruit).."  density: "..string.format("%6d", density).."  windrow"); end;
					fieldData[m].fruit[i].name = thisfruit; fieldData[m].fruit[i].growth[8].density = fieldData[m].fruit[i].growth[8].density + density;
					fieldData[m].fruit[i].growth[8].num = 8;
				end;
			end;
	end;  -- fruitData not nil
	end;  -- numFruits loop

-- Now, check scanData:
-- 1) growth = 1 (growth state 0) and density = 0 means just seeded; use terrain density for fruit density  * UNLESS * there are 2 or more fruits.... then
--		subtract the rest of the fruits' density from terrain density giving the growth state 0's density.
-- 2) if no growth, go thru growth loop 2x2
		local gridDensity = 0;
		local needsPass = false;
			if debug then print(dash75); end;
		for i=1,numFruits do
			if scanData[m].fruit[i].name ~= nil then
				thisfruit = scanData[m].fruit[i].name;
				for o=0, numGrowths-1 do
					if scanData[m].fruit[i].growth[o].num ~= nil then
						if debug then print("        scan  "..string.format("%25s%s%d%s%6d", scanData[m].fruit[i].name, "  [", scanData[m].fruit[i].growth[o].num, "]  ", scanData[m].fruit[i].growth[o].density));  end;
						local growthState = scanData[m].fruit[i].growth[o].num - 1;
						if growthState == 0 and scanData[m].fruit[i].growth[o].density == 0 then
							if debug then print("scanfruits: "..scanfruits);  end;
							if scanfruits > 1 then
								needsPass = true;
							else
								gridDensity = saveDensity;
								scanData[m].fruit[i].growth[o].density = gridDensity;
							end;
						elseif growthState < 0 then  --  Need to go through growth loop 2x2
							local growthCnt = 0; local growthArr = {}; for k=1,5 do growthArr[k]=0; end;
							xx=(x+1024)*2; zz=(z+1024)*2;
							local ww, yy;
							for yy = zz, zz+growthZ, 2 do
								for ww = xx, xx+growthX, 2 do
									growth = getdenregionFunc(fruitData[i].id, ww, yy, 1, 1, 0, 3);
									if growth > 0 and growth < 6 then
										growthArr[growth] = growthArr[growth] + 1;
										growthCnt = growthCnt + 1;
										if debug then print("XX"..string.format("%26s", thisfruit)..string.format("%27s%2d", "growth: ", growth)); end;
									end;
								end;
							end;
							for k=1,5 do
								if growthArr[k] > 0 then
									if debug then print("YY"..string.format("%26s", thisfruit)..string.format("%27s%2d%s%6d", "growth: ", k, "  count: ", growthArr[k])); end;
									scanData[m].fruit[i].growth[k].num = k;
									scanData[m].fruit[i].growth[k].density = mfloor(scanData[m].fruit[i].growth[o].density * growthArr[k] / growthCnt);
									scanData[m].fruit[i].growth[o].num = nil;
								end;
							end;
							if debug then print(string.format("%55s%6d", "growthCnt: ", growthCnt));  end;
						else
							gridDensity = gridDensity + mfloor(scanData[m].fruit[i].growth[o].density / growthState);
						end;
					end;
				end;
			end;
		end;
		if saveDensity ~= 0 then
			if gridDensity ~= saveDensity then  -- go back thru scanData, then growth loop 2x2
				gridDensity = 0;
				for i=1,numFruits do
					if scanData[m].fruit[i].name ~= nil then
						thisfruit = scanData[m].fruit[i].name;
						for o=0, numGrowths-1 do
							if scanData[m].fruit[i].growth[o].num ~= nil then
								if scanData[m].fruit[i].growth[o].density > 0 and scanData[m].fruit[i].growth[o].num < 6 then
									local growthCnt = 0; local growthArr = {}; for k=1,5 do growthArr[k]=0; end;
									xx=(x+1024)*2; zz=(z+1024)*2;
									local ww, yy;
									for yy = zz, zz+growthZ, 2 do
										for ww = xx, xx+growthX, 2 do
											growth = getdenregionFunc(fruitData[i].id, ww, yy, 1, 1, 0, 3);
											if growth > 0 and growth < 6 then
												growthArr[growth] = growthArr[growth] + 1;
												growthCnt = growthCnt + 1;
												if debug then print("X2"..string.format("%26s", thisfruit)..string.format("%27s%2d", "growth: ", growth)); end;
											end;
										end;
									end;
									local scanDensity = scanData[m].fruit[i].growth[o].density;
									for k=1,5 do
										if growthArr[k] > 0 then
											if debug then print("Y2"..string.format("%26s", thisfruit)..string.format("%27s%2d%s%6d", "growth: ", k, "  count: ", growthArr[k])); end;
											scanData[m].fruit[i].growth[o].num = nil;
											scanData[m].fruit[i].growth[k].num = k;
											scanData[m].fruit[i].growth[k].density = mfloor(scanDensity * growthArr[k] / growthCnt);
											if debug then print("        scan 2"..string.format("%25s%s%d%s%6d", scanData[m].fruit[i].name, "  [", scanData[m].fruit[i].growth[k].num, "]  ", scanData[m].fruit[i].growth[k].density));  end;
											local growthState = scanData[m].fruit[i].growth[k].num - 1;
											gridDensity = gridDensity + mfloor(scanData[m].fruit[i].growth[k].density / growthState);
										end;
									end;
									if debug then print(string.format("%55s%6d", "growthCnt: ", growthCnt));  end;
									break;
								end;
							end;
						end;
					end;
				end;
				if needsPass then
					needsPass = false;
					for i=1,numFruits do
						if scanData[m].fruit[i].name ~= nil then
							thisfruit = scanData[m].fruit[i].name;
							for o=0, numGrowths-1 do
								if scanData[m].fruit[i].growth[o].num ~= nil then
									if scanData[m].fruit[i].growth[o].density == 0 then
										scanData[m].fruit[i].growth[o].density = saveDensity - gridDensity;
											if debug then print("using savedensity on: "..thisfruit.."["..o.."]: "..saveDensity-gridDensity);  end;
										gridDensity = gridDensity + scanData[m].fruit[i].growth[o].density;
									end;
								end;
							end;
						end;
					end;
				end;
				if gridDensity ~= saveDensity then  -- must be small differences; use existing values.
					if debug then print("scan data does not match");  end;
					scanNotMatch = scanNotMatch + 1;
					if debug then print("grid: "..gridDensity.."  terr:"..saveDensity);  end;
				end;
			end;
		end;
-- Now, finally, save this grid's fruits & densities in the fieldData array
		for i=1,numFruits do
			if scanData[m].fruit[i].name ~= nil then
				thisfruit = scanData[m].fruit[i].name;
				for o=0, numGrowths-1 do
					if scanData[m].fruit[i].growth[o].num ~= nil then
						if scanData[m].fruit[i].growth[o].density > 0 then
							fieldData[m].fruit[i].name = thisfruit;
							fieldData[m].fruit[i].growth[o].num = o;
							if o > 2 then
								local growthState = scanData[m].fruit[i].growth[o].num - 1;
								fieldData[m].fruit[i].growth[o].density = fieldData[m].fruit[i].growth[o].density + mfloor(scanData[m].fruit[i].growth[o].density / growthState);
							else
								fieldData[m].fruit[i].growth[o].density = fieldData[m].fruit[i].growth[o].density + scanData[m].fruit[i].growth[o].density;
							end;
						end;
					end;
				end;
			end;
		end;

		fieldStatus:reset_scan_data(m)

	end;  -- x (width) loop
	end;  -- z (height) loop
	end;  -- coords not nil
end;  -- maxCoords loop

	if debug then print("scanNotMatch: "..scanNotMatch);  end;

end;  -- fieldData name not nil

-- end;  -- numFields loop
if debug then print("fieldStatus: END func scan_fields @ "..os.date("%H:%M:%S"));  end;
if debug then print(string.format("fieldStatus: elapsed time: %.2f\n", os.clock() - xtime));  end;

end;


function fieldStatus:save_XML()

thisPath = getUserProfileAppPath().."savegame"..g_careerScreen.selectedIndex.."/";
	if debug then print("fieldStatus: trying to create >"..thisPath..fieldDataFile.."<");  end;
fileNew = io.open(thisPath..fieldDataFile, "w");
if fileNew ~= nil then
	fileNew:write("<?xml version=\"1.0\" encoding=\"utf-8\" standalone=\"no\" ?>\n<fieldData>\n");
	fileNew:write(indent.."<map name=\""..thisMap.."\" numFields=\""..numFields.."\" maxCoords=\""..maxCoords.."\">\n");
	fieldStatus:process_fieldData("savexml");
	fileNew:write(indent.."</map>\n</fieldData>\n");
	fileNew:close();
else
	msgTxt = "ERR: can't save to " .. fieldDataFile;
	dispColor = "brightred";
end;

end;


function fieldStatus:process_fieldData(arg)

debug=false; display=false; savexml=false;
if arg == "debug" then debug=true;
elseif arg == "display" then display=true;
elseif arg == "savexml" then savexml=true; end;

if debug then print("BEG func process_fieldData: \"", arg, "\"");  end;

local h, m, n, o, dd, f, bf;
local gNum, gDen, gPct, gState;
local fruitPrinted=false;
local fruitsSeeded;
local plowDensity, cultDensity, seedDensity, fertDensity, fieldArea, fieldDensity, fieldVolume, volumeDiff;
local dispFieldName, dispFruitName;
local fieldHectares, fieldAcres;

isPrinted = false;
hasData = false;
dispDataItems = 0;
dd=1; f=0; h=1;
for m=1,numFields do
	if fieldData[m].name ~= nil then
		dispFieldName = fieldData[m].name;
		fieldArea = fieldData[m].area;
		fieldDensity = fieldArea * densityFactor;
		fieldHectares = fieldData[m].hectares;
		fieldAcres = fieldData[m].acres;

		if debug then print("name: "..dispFieldName.."    area: "..fieldArea); end;
		if savexml then fileNew:write(indent..indent.."<field name=\""..dispFieldName.."\" area=\""..fieldArea.."\" hectares=\""..fieldHectares.."\" acres=\""..fieldAcres.."\">\n"); end;

		dispData[dd].ctrl = "FLD";  dispData[dd].a = dispFieldName; dispData[dd].b = string.format("%4.1f ha / %4.1f ac", fieldHectares, fieldAcres); bf=dd; dd=dd+1; f=f+1;
		if scanField ~= "" and scanFieldArray[1] == dispFieldName then
			start_d = bf;
		end;

		for n=1,maxCoords do
			if fieldData[m].coord[n].startX ~= nil then
				if debug then print("     startX,Z["..n.."]: "..fieldData[m].coord[n].startX..","..fieldData[m].coord[n].startZ.."  endX,Z: "..fieldData[m].coord[n].endX..","..fieldData[m].coord[n].endZ); end;
				if savexml then fileNew:write(indent..indent..indent.."<coordinates startX=\""..fieldData[m].coord[n].startX.."\" startZ=\""..fieldData[m].coord[n].startZ.."\" endX=\""..fieldData[m].coord[n].endX.."\" endZ=\""..fieldData[m].coord[n].endZ.."\" />\n"); end;
			end;
		end;

		fieldVolume = 0; cultDensity = 0; plowDensity = 0; seedDensity = 0; fertDensity = 0;

		for n=0,numStatuses-1 do
			if fieldData[m].status[n].density > 0 then
				if debug then print("     channel["..n.."]: "..chname[n].."  density: "..fieldData[m].status[n].density); end;
				if savexml then fileNew:write(indent..indent..indent.."<status num=\""..n.."\" density=\""..fieldData[m].status[n].density.."\" />\n"); end;
			end;
		end;
		cultDensity=fieldData[m].status[0].density;
		plowDensity=fieldData[m].status[1].density;
		seedDensity=fieldData[m].status[2].density;
		fertDensity=fieldData[m].status[3].density;
		fieldVolume = cultDensity + plowDensity + seedDensity;
			if debug then print(dispFieldName.."  Den: "..fieldDensity.."  Vol: "..fieldVolume);  end;

		fruitsSeeded=false;
		for n=1,FruitUtil.NUM_FRUITTYPES do
			if fieldData[m].fruit[n] ~= nil and fieldData[m].fruit[n].name ~= nil then
				dispFruitName = fieldData[m].fruit[n].name;
				fruitPrinted=false;
				for o=0,numGrowths-1 do
					if fieldData[m].fruit[n].growth[o].num ~= nil and fieldData[m].fruit[n].growth[o].density > 0 then
						gDen=fieldData[m].fruit[n].growth[o].density;
						gPct = (gDen / fieldDensity) * 100;
						if gPct >= 0.1 then
							local e, s = string.find(string.lower(dispFruitName), "manure");    -- save "manure" entries since they are fertilizer
							if e ~= nil then
								holdData[h].ctrl = "FER";  holdData[h].a = dispFruitName;  holdData[h].b = string.format("%3.1f%%", gPct); h=h+1;
							end;
							fruitsSeeded=true;
						end;
						if not fruitPrinted then 
							if debug then print("     fruit: "..dispFruitName); end;
							if savexml then fileNew:write(indent..indent..indent.."<fruit name=\""..dispFruitName.."\">\n"); end;
							fruitPrinted=true;
						end;
						if debug then print("          growth ["..o.."]: "..fieldData[m].fruit[n].growth[o].density); end;
						if savexml then fileNew:write(indent..indent..indent..indent.."<growth num=\""..o.."\" density=\""..fieldData[m].fruit[n].growth[o].density.."\" />\n"); end;
					end;
				end;
				if fruitPrinted then
					if savexml then fileNew:write(indent..indent..indent.."</fruit>\n"); end;
				end;
			end;
		end;

		if seedDensity > 0 and (cultDensity == 0 and plowDensity == 0) and (not fruitsSeeded) then
			dispData[dd].ctrl = "EMP"; dispData[dd].a = "Empty"; dispData[dd].b = "100.0%"; dd=dd+1; f=f+1;
		elseif fieldVolume == 0 and (not fruitsSeeded) then
			dispData[dd].ctrl = "EMP"; dispData[dd].a = "Empty"; dispData[dd].b = "100.0%"; dd=dd+1; f=f+1;
		else
			volumeDiff = (fieldVolume / fieldDensity) * 100;
			if volumeDiff >= 0.1 then
				gPct = 100 - volumeDiff
				dispData[dd].ctrl = "EMP"; dispData[dd].a = "Empty"; dispData[dd].b = string.format("%3.1f%%", gPct); dd=dd+1; f=f+1;
			end;
			if plowDensity > 0 then
				gPct = (plowDensity / fieldDensity) * 100;
				if gPct >= 0.1 then
					dispData[dd].ctrl = "STA";  dispData[dd].a = "Ploughed";  dispData[dd].b = string.format("%3.1f%%", gPct); dd=dd+1; f=f+1;
				end;
			end;
			if cultDensity > 0 then
				gPct = (cultDensity / fieldDensity) * 100;
				if gPct >= 0.1 then
					dispData[dd].ctrl = "STA";  dispData[dd].a = "Cultivated";  dispData[dd].b = string.format("%3.1f%%", gPct); dd=dd+1; f=f+1;
				end;
			end;
			if seedDensity > 0 then
				gPct = (seedDensity / fieldDensity) * 100;
				if gPct >= 0.1 then
					dispData[dd].ctrl = "STA";  dispData[dd].a = "Seeded";  dispData[dd].b = string.format("%3.1f%%", gPct); dd=dd+1; f=f+1;
				end;
			end;
			if fertDensity > 0 then
				gPct = (fertDensity / fieldDensity) * 100;
				if gPct >= 0.1 then
					dispData[dd].ctrl = "FER";  dispData[dd].a = "Fertilized";  dispData[dd].b = string.format("%3.1f%%", gPct); dd=dd+1; f=f+1;
				end;
			end;
			h=1;                                                                                             -- now process any hold entries (manure)
			while true do
				if holdData[h].ctrl == nil then break; end;
				dispData[dd].ctrl = holdData[h].ctrl;
				dispData[dd].a = holdData[h].a;
				dispData[dd].b = holdData[h].b; dd=dd+1; f=f+1;
				h=h+1;
			end;
		end;

-- 2nd pass through fruits
		if fruitsSeeded then
			for n=1,FruitUtil.NUM_FRUITTYPES do
			if fieldData[m].fruit[n] ~= nil and fieldData[m].fruit[n].name ~= nil then
					for o=0,numGrowths-1 do
						if fieldData[m].fruit[n].growth[o].num ~= nil and fieldData[m].fruit[n].growth[o].density > 0 then
							dispFruitName = fieldData[m].fruit[n].name;
							local e, s = string.find(string.lower(dispFruitName), "manure");
							if e == nil then                                                  -- No "manure" found; continue processing
								gNum=fieldData[m].fruit[n].growth[o].num;
								gDen=fieldData[m].fruit[n].growth[o].density;
								gPct = (gDen / fieldDensity) * 100;
								if gPct >= 0.1 then
									if gNum > 0 and gNum < 6 then
										gState = gNum - 1;
										dispFruitName = dispFruitName.." ["..gState.."]";
									elseif gNum == 6 then
										dispFruitName = dispFruitName.." (cutShort)" ;
									elseif gNum == 7 then
										dispFruitName = dispFruitName.." (cutLong)";
									elseif gNum == 8 then
										dispFruitName = dispFruitName.." (windrow)";
									end;
									if gNum > 5 then
										dispData[dd].ctrl = "CUT";
									else
										dispData[dd].ctrl = "FRU";
									end;
									dispData[dd].a = dispFruitName;  dispData[dd].b = string.format("%3.1f%%", gPct); dd=dd+1; f=f+1;
								end;
							end;
						end;
					end;
				end;
			end;
		else  if debug then print("no fruits");  end;
		end;
		if savexml then fileNew:write(indent..indent.."</field>\n"); end;
			if debug then print("f >"..f.."<");  end;
		dispData[bf].ctrl = dispData[bf].ctrl..string.format("NUM%2d", f);
		if f > 2 then
			hasData = true;
		elseif f == 2 and dispData[bf + 1].ctrl ~= "EMP" then
			hasData = true;
		end;
		f=0;
		for m=1,numholdDataEntries do
			holdData[m] = {};
		end;
	end;
end;

dispDataItems = dd - 1;
if debug then fieldStatus:showdispData();  end;
if debug then print("END func process_fieldData: \"", arg, "\"");  end;
debug=false; display=false; savexml=false;

end;


function fieldStatus:showdispData()

local ddd;
print(menuDashLine);
	ddd=1;
	while true do
		if dispData[ddd].ctrl ~= nil then
			if dispData[ddd].a ~= nil then
				if dispData[ddd].b ~= nil then
					print("dispData["..ddd.."] >"..dispData[ddd].ctrl.."|"..dispData[ddd].a.."|"..dispData[ddd].b.."<");
				else
					print("dispData["..ddd.."] >"..dispData[ddd].ctrl.."|"..dispData[ddd].a.."|nil<");
				end;
			else
				print("dispData["..ddd.."] >"..dispData[ddd].ctrl.."|nil<");
			end;
		else
			break;
		end;
		ddd=ddd+1;
	end;
print(menuDashLine);
end;


function fieldStatus:findBottom()

local dLimit, dTemp = 0, 0;
local i, fldCnt = 0, 0;
local bottomFound = false;

dLimit = dispDataItems - dispLinesAvail + 1;
dTemp = dLimit;

while not bottomFound do
	if dispData[dTemp].ctrl ~= "FLD" then  -- Find a 'FLD' line
		for i=dTemp,dispDataItems do
			if string.sub(dispData[i].ctrl, 1, 3) == "FLD" then
				dTemp = i;
				break;
			end;
		end;
	end;
-- Now count the number of fields
	fldCnt = 0;
	for i=dTemp+1,dispDataItems do
		if string.sub(dispData[i].ctrl, 1, 3) == "FLD" then
			fldCnt = fldCnt + 1;
		end;
	end;
	if (dTemp - fldCnt) >= dLimit then
		bottomFound = true;
		start_d = dTemp;
	else						-- Too many fields; find next Field down in the list
		dTemp = dTemp + 1;
	end;
end;

end;


addModEventListener(fieldStatus);
